/* Retro LED Clock
 * http://www.fetchmodus.org/projects/retro
 * Copyright 2016 Nick Ames. Licensed under the MIT license (see end of file).
 * 
 * Displays the time and date using
 * four HP 5082-7300 Dot-matrix numeric displays (time)
 * two  HP 5082-7414 7-segment bubble displays (date)
 *
 * Wiring:
 *   -Hours tens latch (Digit 1): PE2
 *   -Hours ones latch (Digit 2): PD0
 *   -Minutes tens latch (Digit 3): PD4
 *   -Minutes ones latch (Digit 4): PE0
 *   -Bit 1: PB1
 *   -Bit 2: PD6
 *   -Bit 3: PD7
 *   -Bit 4: PB0
 *   -Cathode 1 (leftmost): PD5
 *   -Cathode 2: PB2
 *   -Cathode 3: PE3
 *   -Cathode 4: PC3
 *   -Cathode 5: PC5
 *   -Cathode 6: PD2
 *   -Cathode 7: PD1
 *   -Cathode 8 (rightmost): PD3
 *   -Anode A: PD0
 *   -Anode B: PB0
 *   -Anode C: PB1
 *   -Anode D: PC2
 *   -Anode E: PD7
 *   -Anode F: PC0
 *   -Anode G: PC4
 *   -Anode DP: PC1
 *   
 *   -Colon: PB4
 *   -PM Indicator: PB3
 * 
 *   -Set button: PB5 (Long press starts time setting)
 *   -Up button: PB4 (Press switched between bright/dim.)
 *   -Down button: PB3
 */
#include <avr/interrupt.h>
#define F_CPU 16000000UL
#include <util/delay.h>
#include <util/atomic.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

/* Configuration */
#define USE_AMPM 1 /* If 1, use AM/PM. If 0, use 24-hour display. */
#define SET_PRESS_CYCLES 30 /* Number of 100ms loop iterations that the set button
                             * needs to be held in order to enter time setting mode. */
#define SET_TIMEOUT_CYCLES 300 /* Number of 100ms loop iterations of no buttons being pressed
                                * before time setting mode returns to display mode. */
                                
#define set(reg, bitnum) (reg |= _BV(bitnum))
#define clr(reg, bitnum) (reg &= ~_BV(bitnum))

/* Current display state. */
volatile uint8_t DateDisplay[8]; /* Segment state of date display digits. */
volatile uint8_t TimeDisplay[4]; /* Values of time display digits */
volatile uint8_t PMOn; /* If !0, the PM light show be turned on. */
volatile uint8_t Dim; /* If !0, skip PWM cycles to further dim the display. */

/* Button states
 * Since the colon is shared with button inputs,
 * the buttons can only be read when the colon LEDs
 * are turned off during PWM. This is handled by the
 * colon PWM ISR, with values stored in these globals. */
volatile uint8_t SetPressed, UpPressed, DownPressed; /* Button pressed states. If !0, a button is being held down. */

/* Activate a digit in the seven-segment date display.
 * Digit ranges from 1-8, left-right.
 * Call with digit=0 to deselect all digits. */
void select_date_digit(uint8_t digit){
	clr(DDRD, PD5);
	clr(DDRB, PB2);
	clr(DDRE, PE3);
	clr(DDRC, PC3);
	clr(DDRC, PC5);
	clr(DDRD, PD2);
	clr(DDRD, PD1);
	clr(DDRD, PD3);
	
	switch(digit){
		case 1:
			set(DDRD, PD5);
			break;
		case 2:
			set(DDRB, PB2);
			break;
		case 3:
			set(DDRE, PE3);
			break;
		case 4:
			set(DDRC, PC3);
			break;
		case 5:
			set(DDRC, PC5);
			break;
		case 6:
			set(DDRD, PD2);
			break;
		case 7:
			set(DDRD, PD1);
			break;
		case 8:
			set(DDRD, PD3);
			break;
		default:
			break;
	}
}

/* Set the state of all date display segments.
 * Anode A is bit 0 of the state, B bit 1, ..., G bit 6, DP bit 7. */
void set_date_segments(uint8_t state){
	clr(PORTD, PD6);
	clr(PORTB, PB0);
	clr(PORTB, PB1);
	clr(PORTC, PC2);
	clr(PORTD, PD7);
	clr(PORTC, PC0);
	clr(PORTC, PC4);
	clr(PORTC, PC1);
	if(state & (1 << 0))set(PORTD, PD6);
	if(state & (1 << 1))set(PORTB, PB0);
	if(state & (1 << 2))set(PORTB, PB1);
	if(state & (1 << 3))set(PORTC, PC2);
	if(state & (1 << 4))set(PORTD, PD7);
	if(state & (1 << 5))set(PORTC, PC0);
	if(state & (1 << 6))set(PORTC, PC4);
	if(state & (1 << 7))set(PORTC, PC1);
}

/* Select time digit (1-4, 1 = leftmost). */
void select_time_digit(uint8_t digit){
	PORTD |= _BV(PD4) | _BV(PD0);
	PORTE |= _BV(PE2) | _BV(PE0);
	switch(digit){
		case 1:
			clr(PORTE, PE2);
			break;
		case 2:
			clr(PORTD, PD0);
			break;
		case 3:
			clr(PORTD, PD4);
			break;
		case 4:
			clr(PORTE, PE0);
			break;
	}
}

/* Deselect all time digits. This is required when
 * changing the bitlines, in order to prevent ghosting. */
void deselect_time_digits(void){
	PORTD |= _BV(PD4) | _BV(PD0);
	PORTE |= _BV(PE2) | _BV(PE0);
}

/* Set time bitlines.
 * Note: the bitlines are shared with some segment anodes,
 * so turn off the date cathodes before using this function. */
void set_time_bitlines(uint8_t nybble){
	clr(PORTD, PD6);
	clr(PORTB, PB0);
	clr(PORTB, PB1);
	clr(PORTD, PD7);
	if(nybble & (1 << 1))set(PORTD, PD6);
	if(nybble & (1 << 3))set(PORTB, PB0);
	if(nybble & (1 << 0))set(PORTB, PB1);
	if(nybble & (1 << 2))set(PORTD, PD7);
}

volatile uint8_t CurrentDateDigit;
volatile uint8_t DateStillOn;

/* Display update ISR 1
 * Selects the date digit and turns on the segments. */
ISR(TIMER1_OVF_vect){
	static uint8_t dimcount;
	dimcount++;
	if(!Dim || !(dimcount % 2)){ 
		select_date_digit(0);
		set_time_bitlines(TimeDisplay[0]);
		select_time_digit(1);
		deselect_time_digits();
		set_time_bitlines(TimeDisplay[1]);
		select_time_digit(2);
		deselect_time_digits();
		set_time_bitlines(TimeDisplay[2]);
		select_time_digit(3);
		deselect_time_digits();
		set_time_bitlines(TimeDisplay[3]);
		select_time_digit(4);
		deselect_time_digits();
		
		set_date_segments(0);
		select_date_digit(CurrentDateDigit+1);
		set_date_segments(DateDisplay[CurrentDateDigit]);
		DateStillOn = 1;
		
		CurrentDateDigit++;
		if(CurrentDateDigit > 7){
			CurrentDateDigit = 0;
		}
	}
}

/* Display update ISR 2
 * Turns off the date segments for PWM brightness modulation. */
ISR(TIMER1_COMPA_vect){
	set_date_segments(0);
	DateStillOn = 0;
}

/* Display update ISR 3
 * Turns off time digits for PWM brightness modulation. */
ISR(TIMER1_COMPB_vect){
	select_date_digit(0);
	set_time_bitlines(0x0B); /* Blank */
	PORTD &= ~(_BV(PD4) | _BV(PD0));
	PORTE &= ~(_BV(PE2) | _BV(PE0));
	PORTD |= _BV(PD4) | _BV(PD0);
	PORTE |= _BV(PE2) | _BV(PE0);
	
	if(DateStillOn){
		select_date_digit(CurrentDateDigit+1);
		set_date_segments(DateDisplay[CurrentDateDigit]);
	}
}

/* Timekeeping ISR.
 * Runs at 1Hz; updates current time/date. */
ISR(TIMER0_COMPA_vect){
	static uint8_t count;
	count++;
	if(count == 125){
		count = 0;
		system_tick();
	}
}

/* Colon on ISR. */
ISR(TIMER2_OVF_vect){
	static uint8_t dimcount;
	dimcount++;
	if(!Dim || !(dimcount % 4)){ 
		if(PMOn){
			PORTB &= ~(_BV(PB4) | _BV(PB3));
			DDRB |= _BV(PB4) | _BV(PB3);
		} else {
			PORTB &= ~_BV(PB4);
			DDRB |= _BV(PB4);
		}
	}
	
}

/* Colon off and buttons ISR. */
ISR(TIMER2_COMPA_vect){
	DDRB &= ~(_BV(PB5) | _BV(PB4) | _BV(PB3));
	PORTB |= _BV(PB5) | _BV(PB4) | _BV(PB3); /* Enable pull-ups. */
	asm volatile ("nop\nnop\nnop\nnop");
	SetPressed = !(PINB & _BV(PB5));
	UpPressed = !(PINB & _BV(PB4));
	DownPressed = !(PINB & _BV(PB3));
}


/* Set the displayed value of a date display digit.
 *   digit  is the digit to set (1-8, left-right).
 *   value  is the number to be displayed (0-9).
 *   dp  illuminates the decimal point if !0.
 *   blank  blanks the digit (if dp is not 0, then just the
 *          decimal point will be illuminated). */
void set_date_digit(uint8_t digit, uint8_t value, uint8_t dp, uint8_t blank){
	//TODO: Move this table to flash
	const uint8_t segment_table[] = {
		0b00111111, // 0
		0b00000110, // 1
		0b01011011, // 2
		0b01001111, // 3
		0b01100110, // 4
		0b01101101, // 5
		0b01111101, // 6
		0b00000111, // 7
		0b01111111, // 8
		0b01101111, // 9
	};
	uint8_t val = segment_table[value];
	if(dp){
		val |= 0x80;
	}
	if(blank){
		val = 0;
	}
	DateDisplay[digit - 1] = val;
}

/* Set the displayed value of a time digit.
 *   digit  is the digit to set (1-8, left-right).
 *   value  is the number to be displayed (0-9).
 *   blank  blanks the digit */
void set_time_digit(uint8_t digit, uint8_t value, uint8_t blank){
	if(blank)value = 12;
	TimeDisplay[digit-1] = value;
}

/* Set display brightness to bright mode. */
void bright(void){
	clr(TIMSK1, OCIE1A);
	clr(TIMSK1, OCIE1B);
	OCR2A = 245;
	Dim = 0;
}

/* Set display brightness to dim mode. */
void dim(void){
	OCR1A = 800; /* Date display brightness level. */
	OCR1B = 500; /* Time display brightness level. */
	set(TIMSK1, OCIE1A);
	set(TIMSK1, OCIE1B);
	OCR2A = 5;
	Dim = 1;
}

/* Set colon to AM mode. */
void set_am(void){
	PMOn = 0;
}

/* Set colon to PM mode. */
void set_pm(void){
	PMOn = 1;
}

/* Initialize peripherals. */
void init(void){
	/* Date display digits (cathodes). 
	 * These are switched between output and High-Z
	 * by select_data_digit(). */
	clr(PORTD, PD5);
	clr(PORTB, PB2);
	clr(PORTE, PE3);
	clr(PORTC, PC3);
	clr(PORTC, PC5);
	clr(PORTD, PD2);
	clr(PORTD, PD1);
	clr(PORTD, PD3);
	
 	/* Date display segments (anodes). */
	set(DDRD, PD6);
	set(DDRB, PB0);
	set(DDRB, PB1);
	set(DDRC, PC2);
	set(DDRD, PD7);
	set(DDRC, PC0);
	set(DDRC, PC4);
	set(DDRC, PC1);
 	
	/* Time display digit selects. */
	set(DDRE, PE2);
	set(DDRD, PD0);
	set(DDRD, PD4);
	set(DDRE, PE0);

	/* Timer 0 drives the timekeeping ISR at 125Hz. */
	TCCR0A = _BV(WGM01); /* Mode: CTC (top=OCRA) */
	TCCR0B = _BV(CS02) | _BV(CS00);
	OCR0A = 124;
	TIMSK0 = _BV(OCIE0A); /* Enable OCRA interrupt. */

	/* Timer 1 drives the date display scanning ISR at ~1kHz. */
	TCCR1A = 0; /* Mode: phase and frequency correct PWM (WGM13 = 1) (top = ICR1) */
	TCCR1B = _BV(CS10) | _BV(WGM13); /* CLKio/1 */
	TIMSK1 = _BV(OCIE1B) | _BV(OCIE1A) | _BV(TOIE1);
	uint16_t icrval = 16000;
	ICR1H = icrval >> 8;
	ICR1L = icrval & 0xFF;
	
	
	/* TODO: Timer 2 for colon PWM. */
	TCCR2B = _BV(CS22); /* Set normal mode, ~1kHz interrupt rate. */
	TIMSK2 = _BV(TOIE2) | _BV(OCIE2A);
	
	bright();
	
	/* Set button
	 * (other buttons handled by colon PWM ISR. */
	PORTB |= _BV(PB5); /* Input, pull-up activated. */
	
	/* Set Initial Time */
	set_system_time(1476323709l - 946684800l);
	set_zone(-7 * ONE_HOUR);
	sei();
}

int main(void){
	init();
	
	enum {UNSET, DISPLAY, SET_YEAR, SET_MONTH, SET_DAY, SET_HOUR, SET_MINUTE, SET_TIME} state = UNSET;
	uint8_t minute = 0;
	uint8_t hour = 12;
	uint16_t year = 2016;
	uint16_t month = 1;
	uint16_t day = 1;
	uint8_t set_prev = SetPressed;
	uint8_t up_prev = UpPressed;
	uint8_t down_prev = DownPressed;
	uint16_t cycles_set_pressed = 0; /* Number of loop iteration during which set has been held. */
	uint16_t cycles_since_last_pressed = 0; /* Number of iterations that no buttons have been pressed. */
	uint8_t blink_count = 0;
	struct {
		unsigned int year : 1;
		unsigned int month : 1; 
		unsigned int day : 1;
		unsigned int hour : 1;
		unsigned int minute : 1;
	} blink; /* Fields to blink, if the corresponding bit is 1. */
		
	/* The main loop goes around at approx. 10 hz. */
	while(1){
		if(SetPressed){
			cycles_set_pressed++;
		} else {
			cycles_set_pressed = 0;
		}
		if(SetPressed || UpPressed || DownPressed){
			cycles_since_last_pressed = 0;
		} else {
			cycles_since_last_pressed++;
		}
		if(SET_TIME == state){
			ATOMIC_BLOCK(ATOMIC_RESTORESTATE){
				struct tm cur_tm;
				cur_tm.tm_sec  = 0;
				cur_tm.tm_min  = minute;
				cur_tm.tm_hour = hour;
				cur_tm.tm_mday = day;
				cur_tm.tm_mon  = month - 1;
				cur_tm.tm_year  = year - 1900;
				cur_tm.tm_isdst = 0;
				set_system_time(mktime(&cur_tm));
			}
			state = DISPLAY;
		} else if(DISPLAY == state){
			struct tm *cur_tm;
			time_t cur_time = time(NULL);
			cur_tm = localtime(&cur_time);
			minute = cur_tm->tm_min;
			hour = cur_tm->tm_hour;
			year = cur_tm->tm_year + 1900;
			month = cur_tm->tm_mon + 1;
			day = cur_tm->tm_mday;
			blink.year = 0;
			blink.month = 0;
			blink.day = 0;
			blink.hour = 0;
			blink.minute = 0;
			if(UpPressed && !up_prev){
				if(Dim)bright();
				else dim();
			}
			if(cycles_set_pressed > SET_PRESS_CYCLES){
				state = SET_YEAR;
			}
		} else if(UNSET == state){
			blink.year = 1;
			blink.month = 1;
			blink.day = 1;
			blink.hour = 1;
			blink.minute = 1;
			if(cycles_set_pressed > SET_PRESS_CYCLES){
				state = SET_YEAR;
			}
		} else if(SET_YEAR == state){
			bright();
			blink.year = 1;
			blink.month = 0;
			blink.day = 0;
			blink.hour = 0;
			blink.minute = 0;
			if(UpPressed && !up_prev){
				year++;
			}
			if(DownPressed && !down_prev){
				year--;
			}
			if(SetPressed && !set_prev){
				state = SET_MONTH;
			}
			if(cycles_since_last_pressed > SET_TIMEOUT_CYCLES){
				state = SET_TIME;
			}
		} else if(SET_MONTH == state){
			blink.year = 0;
			blink.month = 1;
			blink.day = 0;
			blink.hour = 0;
			blink.minute = 0;
			if(UpPressed && !up_prev && month < 12){
				month++;
			}
			if(DownPressed && !down_prev && month > 1){
				month--;
			}
			if(SetPressed && !set_prev){
				state = SET_DAY;
			}
			if(cycles_since_last_pressed > SET_TIMEOUT_CYCLES){
				state = SET_TIME;
			}
		} else if(SET_DAY == state){
			blink.year = 0;
			blink.month = 0;
			blink.day = 1;
			blink.hour = 0;
			blink.minute = 0;
			if(UpPressed && !up_prev && day < 31){
				day++;
			}
			if(DownPressed && !down_prev && day > 1){
				day--;
			}
			if(SetPressed && !set_prev){
				state = SET_HOUR;
			}
			if(cycles_since_last_pressed > SET_TIMEOUT_CYCLES){
				state = SET_TIME;
			}
		} else if(SET_HOUR == state){
			blink.year = 0;
			blink.month = 0;
			blink.day = 0;
			blink.hour = 1;
			blink.minute = 0;
			if(UpPressed && !up_prev && hour < 23){
				hour++;
			}
			if(DownPressed && !down_prev && hour > 0){
				hour--;
			}
			if(SetPressed && !set_prev){
				state = SET_MINUTE;
			}
			if(cycles_since_last_pressed > SET_TIMEOUT_CYCLES){
				state = SET_TIME;
			}
		} else if(SET_MINUTE == state){
			blink.year = 0;
			blink.month = 0;
			blink.day = 0;
			blink.hour = 0;
			blink.minute = 1;
			if(UpPressed && !up_prev && minute < 59){
				minute++;
			}
			if(DownPressed && !down_prev && minute > 0){
				minute--;
			}
			if(SetPressed && !set_prev){
				state = SET_TIME;
			}
			if(cycles_since_last_pressed > SET_TIMEOUT_CYCLES){
				state = SET_TIME;
			}
		} 
		
		uint8_t disp_hour = hour;
		blink_count++;
		if(USE_AMPM){
			if(disp_hour > 11){
				disp_hour = disp_hour - 12;
				set_pm();
			} else {
				set_am();
			}
			if(disp_hour == 0)disp_hour = 12;
		} else {
			set_am();
		}
		set_time_digit(1, disp_hour/10, disp_hour/10 == 0 || (blink.hour && (blink_count % 2)));
		set_time_digit(2, disp_hour % 10, (blink.hour && (blink_count % 2)));
		set_time_digit(3, minute/10, (blink.minute && (blink_count % 2)));
		set_time_digit(4, minute%10, (blink.minute && (blink_count % 2)));
		
		set_date_digit(1, year/1000, 0, (blink.year && (blink_count % 2)));
		set_date_digit(2, (year/100) % 10, 0, (blink.year && (blink_count % 2)));
		set_date_digit(3, (year/10) % 10, 0, (blink.year && (blink_count % 2)));
		set_date_digit(4, year% 10, 1, (blink.year && (blink_count % 2)));
		set_date_digit(5, month/10, 0, (blink.month && (blink_count % 2)));
		set_date_digit(6, month%10, 1, (blink.month && (blink_count % 2)));
		set_date_digit(7, day/10, 0, (blink.day && (blink_count % 2)));
		set_date_digit(8, day%10, 0, (blink.day && (blink_count % 2)));
		
		set_prev = SetPressed;
		up_prev = UpPressed;
		down_prev = DownPressed;
		_delay_ms(100);
	}
	return(0);
}


/* License:
 * MIT License
 * 
 * Copyright (c) 2016 Nick Ames
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */