/* Retro LED Clock 3D-Printed Part
 * http://www.fetchmodus.org/projects/retro/
 * Copyright 2016 Nick Ames. Licensed Creative Commons Zero (public domain).
 * 
 * This file contains three objects:
 *   inner_holder() - a 3D-printed part that holds the PCB and other components
 *                    of the clock inside the case.
 *   case() - a model of the case, which is made from wood.
 *   case_back() - a model of the case back, which is made from wood.
 */

/* Configuration (all distances in mm) */
CASE = [59, 49.5, 43]; /* Case outer dimensions. */
WINDOWHOLE = [46, 30.5]; /* Size of window hole in case [X,Z]. */
CAVITY = [49, 46, 33.5]; /* Size of inside of case. */
BACKTHICK = 3.6; /* Thickness of case back. */

WINDOW = [48.3, 32.5];
WINDOWTHICK = 3;
LEDHEIGHT = 9.2; /* Height of top of LED displays above PCB. */
PCBTHICK = 1.6;
PCB = [45.5, 31];
HOLE_UP = 7.62;/* Distance of holes from bottom of PCB. */
HOLE_RIGHT = [3.66, 39.92]; /* Distance of holes from left edge of PCB. */
HOLE_DIA = 2.8;
HOLE_DEPTH = 7;
HOLDER = [48.2, CAVITY[1] - WINDOWTHICK  - BACKTHICK - 1, 33.2]; /* Size of 3D printed internal piece. */

BUTTONS_V = -7; /* Vertical position of buttons, from center of case. */
POWER_X = -13; /* Horizontal position of power jack, from center of case. */
BUTTONS_CLUSTER_X = 7; /* Horizontal position of button cluster, from center of case. */
BUTTONS_SPACING = 8; /* Horizontal space between buttons. */
BUTTONS_Z = CASE[2]/2 + BUTTONS_V;
BUTTONS_X = [BUTTONS_CLUSTER_X - BUTTONS_SPACING, BUTTONS_CLUSTER_X, BUTTONS_CLUSTER_X + BUTTONS_SPACING];

module case(){
	difference(){
		translate([-CASE[0]/2, 0, 0])cube(CASE);
		translate([0, 0, CASE[2]/2])cube([WINDOWHOLE[0], 100, WINDOWHOLE[1]], center=true);
		translate([-CAVITY[0]/2, CASE[1]-CAVITY[1], (CASE[2]-CAVITY[2])/2])cube([CAVITY[0], CAVITY[1] + 2, CAVITY[2]]);
	}
}

module case_back(){
	translate([-CAVITY[0]/2, CASE[1]-BACKTHICK-.5, (CASE[2]-CAVITY[2])/2])cube([CAVITY[0], BACKTHICK, CAVITY[2]]);
}

/* Cut-out for power jack, with axis on y+ and edge on y=0. */
module power_jack(){
	w = 9.5;
	zo = 4.4;
	d = 14.7;
	translate([-w/2, -d, -zo])cube([w, d, 100]);
	rotate([-90, 0, 0])cylinder(r=1, h=10);
}

/* Button cut-out, with axis on y+ and edge on y=0. */
module button(){
	w = 6.5;
	h = 8.7;
	d = 10;
	bh = 4;
	bp = 3.7; 
	difference(){
		translate([-w/2, -d, -h/2])cube([w, d, h]);
		translate([-10, -d, -bh/2])cube([20, bh, bp]);
	}
}

module inner_holder(){
	psx = PCB[0] - 2;
	psz = PCB[1] - 2;
	pcbo = [-PCB[0]/2, CASE[1] - CAVITY[1] + WINDOWTHICK + LEDHEIGHT, (CASE[2]-PCB[1])/2];
	difference(){
		translate([-HOLDER[0]/2, CASE[1] - CAVITY[1], (CASE[2]-HOLDER[2])/2])cube(HOLDER);
		translate([-WINDOW[0]/2, CASE[1] - CAVITY[1], (CASE[2]-WINDOW[1])/2])cube([WINDOW[0], WINDOWTHICK, WINDOW[1]]);
		translate([-PCB[0]/2, CASE[1] - CAVITY[1] + WINDOWTHICK, (CASE[2]-PCB[1])/2])cube([PCB[0], PCBTHICK + LEDHEIGHT, PCB[1]]);
		translate([-psx/2, CASE[1] - CAVITY[1] + WINDOWTHICK + PCBTHICK + LEDHEIGHT, (CASE[2]-psz)/2])cube([psx, 3, psz]);
		translate([pcbo[0] + 27, 0, pcbo[2] + 18])cube([18.5, 100, 13]);
		translate([pcbo[0] + 27, CASE[1] - CAVITY[1] + WINDOWTHICK + PCBTHICK + LEDHEIGHT, pcbo[2] + 18])cube([18.5, 100, 13]);
		translate([pcbo[0] + HOLE_RIGHT[0], pcbo[1], pcbo[2] + HOLE_UP])rotate([-90, 0, 0])cylinder(d=HOLE_DIA, h=HOLE_DEPTH + PCBTHICK);
		translate([pcbo[0] + HOLE_RIGHT[1], pcbo[1], pcbo[2] + HOLE_UP])rotate([-90, 0, 0])cylinder(d=HOLE_DIA, h=HOLE_DEPTH + PCBTHICK);
		translate([pcbo[0] + 27, CASE[1] - CAVITY[1] + WINDOWTHICK + PCBTHICK + LEDHEIGHT + 4, pcbo[2] + 18])cube([20, 100, 15]);
		translate([POWER_X, CASE[1] - CAVITY[1] + HOLDER[1], BUTTONS_Z])power_jack();
		for(x = BUTTONS_X){
			translate([x, CASE[1] - CAVITY[1] + HOLDER[1], BUTTONS_Z])button();
		}
		translate([-10, 29, 32])cube([150, 4, 100]);
		translate([BUTTONS_X[1]-1.5, CASE[1] - CAVITY[1] + HOLDER[1] - 2, 15])cube([3, 10, 10]);
		translate([BUTTONS_X[1]-7.5, CASE[1] - CAVITY[1] + HOLDER[1] - 2, 16])cube([15, 10, 3]);
	}
}

module drill_guide(){
	my = 3;
	mx = 13;
	w = CAVITY[0] + 2*mx;
	h = CAVITY[2] + 2*my;
	d = 10;
	difference(){
		translate([-w/2, -my, 0])cube([w, h, d]);
		translate([-CAVITY[0]/2, 0, 0])cube([CAVITY[0], CAVITY[2] + .75, BACKTHICK - .5]);
		translate([POWER_X, CAVITY[2]/2 + BUTTONS_V, -1])cylinder(d = 9.5, h = 100);
		for(x = BUTTONS_X){
			translate([x, CAVITY[2]/2 + BUTTONS_V, -1])cylinder(d = 3.7, h = 100);
		}
		translate([CAVITY[0]/2 + 5, h/2 - my, -1])cylinder(d=5, h = 100);
		translate([-CAVITY[0]/2 - 5, h/2 - my, -1])cylinder(d=5, h = 100);
	}
}

//rotate([180, 0, 0])drill_guide();
rotate([-90, 0, 0])inner_holder();
//button();
//case();
